<?php

/**
 * @package     Extly.Components
 * @subpackage  com_xtdir - Extended Directory for SobiPro
 *
 * @author      Prieco S.A. <support@extly.com>
 * @copyright   Copyright (C) 2007 - 2014 Prieco, S.A. All rights reserved.
 * @license     http://www.gnu.org/licenses/gpl-3.0.html GNU/GPL
 * @link        http://www.extly.com http://support.extly.com
 */
// No direct access
defined('_JEXEC') or die('Restricted access');

/**
 * TreeIndexer.
 *
 * @package     Extly.Components
 * @subpackage  com_xtdir
 * @since       1.0
 */
class TreeIndexer
{
	private $catalog_list;

	private $index_limit;

	private $dbproxy;

	private $db;

	/**
	 * Constructor.
	 *
	 */
	public function __construct()
	{
		$this->catalog_list = EExtensionHelper::getExtensionParam('system', 'xtdir', 'catalog_list');
		$this->catalog_list = $this->_filterCatalogs();
		$this->index_limit = EExtensionHelper::getExtensionParam('system', 'xtdir', 'index_limit', 1000);
		$this->_in_memory = EExtensionHelper::getExtensionParam('system', 'xtdir', 'in_memory', 0);
		$this->logger = XtdirLogger::getInstance();
		$this->db = JFactory::getDbo();
		$this->dbproxy = EDbProxyHelper::getInstance();
	}

	/**
	 * Check last run
	 *
	 * @param   bool  $force  Force to index
	 *
	 * @return	bool
	 */
	public function indexSinceLastRunRebuild($force = false)
	{
		if (($force) || ($this->_checkLastRun()))
		{
			$this->logger->log(JLog::INFO, 'indexSinceLastRunRebuild ' . $force);
			$catalogs = $this->catalog_list;
			$this->logger->log(JLog::INFO, 'indexSinceLastRunRebuild Catalogs ', $catalogs);

			if (!empty($catalogs))
			{
				$this->_rebuild();
			}
		}

		return true;
	}

	/**
	 * _rebuild tree and associated structures
	 *
	 * @return	bool
	 */
	private function _rebuild()
	{
		$this->rebuildInitialization();

		if (defined('XTDIR_CRONJOB_RUNNING'))
		{
			echo JProfiler::getInstance()->mark('Building catalogs') . "\n";
		}

		$this->logger->log(JLog::INFO, 'Building catalogs');

		// Rebuilding
		$catalogs = $this->catalog_list;

		foreach ($catalogs as $catalog_pid)
		{
			if (!$this->rebuildCatalog($catalog_pid))
			{
				return false;
			}
		}

		$this->rebuildFinalization();

		return true;
	}

	/**
	 * rebuildInitialization
	 *
	 * @return	void
	 */
	public function rebuildInitialization()
	{
		if (defined('XTDIR_CRONJOB_RUNNING'))
		{
			echo JProfiler::getInstance()->mark('Cleaning catalogs') . "\n";
		}

		$this->logger->log(JLog::INFO, 'Cleaning catalogs');

		$this->_createTempTables();

		// Also Stats
		$statsHelper = new StatsHelper;
		$statsHelper->cleanTotals();
	}

	/**
	 * rebuildCatalog
	 *
	 * @param   int   $catalog_pid  Param
	 * @param   bool  $continue     Param
	 *
	 * @return	bool
	 */
	public function rebuildCatalog($catalog_pid, $continue = false)
	{
		// Main Index Tree
		if (defined('XTDIR_CRONJOB_RUNNING'))
		{
			echo JProfiler::getInstance()->mark('Rebuilding tree: ' . $catalog_pid) . "\n";
		}

		$this->logger->log(JLog::INFO, 'Rebuilding tree: ' . $catalog_pid);

		$result = $this->_rebuild_tree($catalog_pid, $continue);
		$this->dbproxy->flushInserts();

		return $result;
	}

	/**
	 * rebuildFinalization
	 *
	 * @return	void
	 */
	public function rebuildFinalization()
	{
		// Cleaning
		if (defined('XTDIR_CRONJOB_RUNNING'))
		{
			echo JProfiler::getInstance()->mark('Cleaning and publishing new core index') . "\n";
		}

		$this->logger->log(JLog::INFO, 'Cleaning and publishing new core index');
		$this->_clean_tree();

		// Calculating totals
		$statsHelper = new StatsHelper;
		$catalogs = $this->catalog_list;

		foreach ($catalogs as $catalog_pid)
		{
			if (defined('XTDIR_CRONJOB_RUNNING'))
			{
				echo JProfiler::getInstance()->mark('Calculating totals: ' . $catalog_pid) . "\n";
			}

			$this->logger->log(JLog::INFO, 'Calculating totals: ' . $catalog_pid);
			$statsHelper->calculateTotals($catalog_pid);
		}

		// Also Promoting entries
		if (XTDIR_PROMOTED_ENTRIES)
		{
			if (defined('XTDIR_CRONJOB_RUNNING'))
			{
				echo JProfiler::getInstance()->mark('Promoting entries') . "\n";
			}

			$this->logger->log(JLog::INFO, 'Promoting entries');
			$promotedHelper = new PromotedHelper;
			$promotedHelper->index();
		}

		$this->_dropTempTables();
	}

	/**
	 * Rebuild_tree
	 *
	 * @param   object  $catalog_id  Catalog
	 * @param   bool    $continue    Param
	 *
	 * @return	bool
	 */
	private function _rebuild_tree($catalog_id, $continue = false)
	{
		/*
		 * Storing Hierarchical Data in a Database Article Modified Preorder Tree Traversal Script that converts an adjacency list to a modified preorder tree traversal table Based on http://www.sitepoint.com/hierarchical-data-database/
		 */
		$this->logger->log(JLog::INFO, "Rebuilding tree", $catalog_id);

		$max_execution = EExtensionHelper::getExtensionParam('system', 'xtdir', 'max_execution', 10);

		if (defined('XTDIR_CRONJOB_RUNNING'))
		{
			$max_execution = 0;
		}

		$date = new DateTime;
		$timestamp = $date->getTimestamp();

		// State Initialization
		if ($continue)
		{
			$state = self::_loadState();

			$this->logger->log(JLog::INFO, '_rebuild_tree - continue ', $state);

			$data = json_decode($state);
			$cursor = $data->cursor;
			$current_node = $data->current_node;
		}
		else
		{
			$this->logger->log(JLog::INFO, "Rebuilding tree - Start", $catalog_id);
			$cursor = 1;
			$current_node = new StdClass;
			$current_node->id = $catalog_id;
			$current_node->left = $cursor;
			$current_node->parent = null;
			$current_node->path = '';

			self::_saveState('');
		}

		while ($current_node)
		{
			$date = new DateTime;
			$now = $date->getTimestamp();

			if ((!defined('XTDIR_CRONJOB_RUNNING')) && ($max_execution) && (($now - $timestamp) > $max_execution))
			{
				$state = compact('cursor', 'current_node');
				$data = json_encode($state);

				$this->logger->log(JLog::INFO, "Rebuilding tree - max_execution!", $data);

				self::_saveState($data);

				return false;
			}

			if (isset($current_node->node_ids))
			{
				$node_ids = $current_node->node_ids;
			}
			else
			{
				$node_ids = $this->_getChildren($current_node->id);
			}

			if (empty($node_ids))
			{
				$cursor++;
				$current_node->right = $cursor;

				if (isset($current_node->parent))
				{
					$this->_insertNode($current_node->id, $current_node->parent->id, $current_node->left, $current_node->right, $current_node->path, $catalog_id);

					// Safety Limit
					if (($this->index_limit) && (($cursor / 2) > $this->index_limit))
					{
						return true;
					}
				}

				$current_node = $current_node->parent;
			}
			else
			{
				$next_id = array_pop($node_ids);
				$current_node->node_ids = $node_ids;

				$cursor++;

				$next_node = new StdClass;
				$next_node->id = $next_id;
				$next_node->left = $cursor;
				$next_node->parent = $current_node;
				$next_node->path = $current_node->path . $current_node->id . '-';

				$current_node = $next_node;
			}
		}

		return true;
	}

	/**
	 * Insert a node
	 *
	 * @param   int     $node        Current node
	 * @param   int     $parent      Parent node
	 * @param   int     $left        Left order
	 * @param   int     $right       Right order
	 * @param   string  $path        Path
	 * @param   int     $catalog_id  Catalog
	 *
	 * @return	void
	 */
	private function _insertNode($node, $parent, $left, $right, $path, $catalog_id)
	{
		$now = JFactory::getDate();
		$nowSql = $now->toSql();

		// Inserted as negative values to be disabled, a quick way to make staged an index while it's generated
		$values = array(
			$node * -1,
			$parent * -1,
			$left,
			$right,
			$this->db->Quote($path),
			$catalog_id * -1,
			$this->db->Quote($nowSql)
		);
		$this->dbproxy->addInsert('#__xtdir_trees', $values);
	}

	/**
	 * Get all children of this node
	 *
	 * @param   object  $node  Current node
	 *
	 * @return	array()
	 */
	private function _getChildren($node)
	{
		$query = $this->db->getQuery(true);

		$query->select('id');
		$query->from($this->db->quoteName('#__tmp_object_relations'));
		$query->where('pid = ' . $this->db->Quote($node));

		// Order
		$query->order('id desc');

		// Safety Limit
		if ($this->index_limit)
		{
			$query = $query->__toString() . ' LIMIT ' . $this->index_limit;
		}

		$this->db->setQuery($query);
		$rows = $this->db->loadColumn();

		return $rows;
	}

	/**
	 * index.
	 *
	 * @param   bool  $force  Param
	 *
	 * @return	void
	 */
	public static function index($force = false)
	{
		$logger = XtdirLogger::getInstance();
		$logger->log(JLog::INFO, "index", $force);

		$cronjob_mode = EExtensionHelper::getExtensionParam('system', 'xtdir', 'cronjob_mode', 1);

		if ((!$cronjob_mode) || ($force))
		{
			$scheduled_indexer = new TreeIndexer;
			$scheduled_indexer->indexSinceLastRunRebuild($force);
		}
	}

	/**
	 * _loadState.
	 *
	 * @return	string
	 */
	private static function _loadState()
	{
		$state = EExtensionHelper::getExtensionParam('system', 'xtdir', 'indexState');

		return $state;
	}

	/**
	 * _saveState.
	 *
	 * @param   string  $data  Param
	 *
	 * @return	bool
	 */
	private static function _saveState($data)
	{
		return EExtensionHelper::setExtensionParam('system', 'xtdir', 'indexState', $data);
	}

	/**
	 * getCatalogs.
	 *
	 * @return	array
	 */
	private function _filterCatalogs()
	{
		$catalog_list = $this->catalog_list;

		if (!empty($catalog_list))
		{
			$catalogs = $catalog_list;
		}
		else
		{
			$catalogs = SectionsHelper::findAll();
		}

		if ((!XTDIR_MULTI_MODE) && (count($catalogs) > 1))
		{
			array_splice($catalogs, 1);
		}

		return $catalogs;
	}

	/**
	 * Clean the tree
	 *
	 * @return	void
	 */
	private function _clean_tree()
	{
		$this->logger->log(JLog::INFO, 'Cleaning tree');

		// Inserted as negative values to be disabled, a quick way to make staged an index while it's generated
		$query = 'DELETE FROM ' . $this->db->quoteName('#__xtdir_trees') . ' WHERE id > 0;';
		$this->db->setQuery($query);
		$this->db->execute();

		$query = 'UPDATE ' . $this->db->quoteName('#__xtdir_trees') . ' SET id = id * -1, parent = parent * -1, section = section * -1;';
		$this->db->setQuery($query);
		$this->db->execute();
	}

	/**
	 * Check last run internal
	 *
	 * @return	void
	 */
	private function _checkLastRun()
	{
		// Get a reference to the global cache object.
		$config = JFactory::getConfig();
		$cache_enabled = $config->get('caching');

		$this->logger->log(JLog::INFO, '_checkLastRun - Cache ' . $cache_enabled);

		if (!$cache_enabled)
		{
			return true;
		}

		$query = "SELECT last_update FROM " . $this->db->quoteName('#__xtdir_trees') . " LIMIT 1";
		$this->db->setQuery($query);
		$last_run = $this->db->loadResult();
		$last_run = strtotime($last_run);

		$interval = $config->get('cachetime');

		$this->logger->log(JLog::INFO, '_checkLastRun - Index Last Run ' . $last_run);
		$this->logger->log(JLog::INFO, '_checkLastRun - Interval ' . $interval);

		return (time() > $last_run + $interval * 60);
	}

	/**
	 * Get catalogs
	 *
	 * @return	array
	 */
	public function getCatalogs()
	{
		return $this->catalog_list;
	}

	/**
	 * getTotal
	 *
	 * @param   int  $section_id  Section
	 *
	 * @return	int
	 */
	public function getTotal($section_id)
	{
		$query = "SELECT COUNT(*) FROM " . $this->db->quoteName('#__xtdir_trees') . " WHERE " . $this->db->quoteName('section') . " = " . $this->db->q($section_id * -1);
		$this->db->setQuery($query);
		$total = $this->db->loadResult();

		return $total;
	}

	/**
	 * _createTempTables
	 *
	 * @return	void
	 */
	private function _createTempTables()
	{
		$query = "DROP TABLE IF EXISTS `#__tmp_object_relations`;";
		$this->db->setQuery($query);
		$this->db->execute();

		$query = "CREATE " . ($this->_in_memory ? 'TEMPORARY' : '') . " TABLE IF NOT EXISTS `#__tmp_object_relations` (
	`id`         INT(11)     NOT NULL DEFAULT '0',
	`pid`        INT(11)     NOT NULL DEFAULT '0',
	KEY `id` (`id`),
	KEY `pid` (`pid`)
)
DEFAULT CHARSET = utf8 " . ($this->_in_memory ? 'ENGINE=MEMORY' : '') . ";";
		$this->db->setQuery($query);
		$this->db->execute();

		$now = JFactory::getDate();
		$nowSql = $now->toSql();

		$query = "INSERT INTO `#__tmp_object_relations`
					SELECT r.id, r.pid
						FROM `#__sobipro_relations` AS r,
							 `#__sobipro_object` AS o
						WHERE r.id = o.id
								AND o.state = 1
								AND o.approved = 1
								AND (o.validUntil = 0 OR o.validUntil > " . $this->db->Quote($nowSql) . ')';

		$this->db->setQuery($query);
		$this->db->execute();
	}

	/**
	 * _dropTempTables
	 *
	 * @return	void
	 */
	private function _dropTempTables()
	{
		$query = "DROP TABLE IF EXISTS `#__tmp_object_relations`;";
		$this->db->setQuery($query);
		$this->db->execute();
	}
}
